/*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		IP/TCP/UDP checksumming routines
 *
 * Authors:	Jorge Cwik, <jorge@laser.satlink.net>
 *		Arnt Gulbrandsen, <agulbra@nvg.unit.no>
 *		Tom May, <ftom@netcom.com>
 *              Pentium Pro/II routines:
 *              Alexander Kjeldaas <astor@guardian.no>
 *              Finn Arne Gangstad <finnag@guardian.no>
 *		Lots of code moved from tcp.c and ip.c; see those files
 *		for more names.
 *
 * Changes:     Ingo Molnar, converted csum_partial_copy() to 2.1 exception
 *			     handling.
 *		Andi Kleen,  add zeroing on error converted to pure assembler
 *		2002-10-30 Denis Vlasenko
 *			boot-time benchmarking, 3Dnow/MMX+/SSE versions
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */

/*
** computes a partial checksum, e.g. for TCP/UDP fragments
**
** unsigned int csum_partial(const unsigned char * buff,
**		int len, unsigned int sum)
*/
	   
.text
.align 4
.globl csum_partial

csum_partial:
	pushl	%esi
	pushl	%ebx
	movl	20(%esp), %eax	# arg: sum
	movl	16(%esp), %ecx	# arg: len
	movl	12(%esp), %esi	# arg:	buf

	testl	$3, %esi
	jz	40f
20:
    # not 4-aligned: analyze and align...
	testl	$1, %esi
	jz	30f
	
    # unaligned start addr
	decl	%ecx
	js	90f 		# sz==0, exit
	movzbl	(%esi), %ebx	# eat one byte...
	addl	%ebx, %eax
	adcl	$0, %eax
	roll	$8, %eax	# NB: need to be undone at exit!
	incl	%esi
	testl	$2, %esi
	jz	40f
30:
    # 2-aligned, but not 4-aligned
	cmpl	$3, %ecx
	jbe	60f
	addw	(%esi), %ax	# eat 2 bytes
	leal	2(%esi), %esi
	adcl	$0, %eax
	subl	$2, %ecx
40:
    # esi is 4-aligned here, call block routine
	movl	$csum_basic, %ebx	# known ok even for ecx==0 etc
	cmpl	$128, %ecx		# use optimized routine
	jb	50f			#   only for large blocks
	movl	best_csum, %ebx
50:	call	*%ebx
60:
    # handle the last 0-3 bytes without much jumping
	jecxz	80f
	notl	%ecx		# 0->3, 1->2, 2->1, 3->0, higher bits are masked
	movl	$0xffffff, %ebx	#   by the shll and shrl instructions
	shll	$3, %ecx
	shrl	%cl, %ebx
	andl	(%esi), %ebx	# esi is 4-aligned so should be ok
	addl	%ebx, %eax
	adcl	$0, %eax
80:
    # undo csum rotation if start addr was odd
	testl	$1, 12(%esp)
	jz	90f
	roll	$8, %eax
90: 
	popl	%ebx
	popl	%esi
	ret
