#include <stdio.h>
#include <stdlib.h>

//Intel
#define PREFETCH " prefetchnta "
#define FEMMS ""
#define EMMS ""
#define SFENCE " sfence\n"

//AMD
//#define PREFETCH " prefetch "
//#define FEMMS " femms\n"
//#define EMMS " emms\n"
//#define SFENCE " sfence\n"

#define NAME(a) \
unsigned int csum_##a(const unsigned char * buff, int len, \
			unsigned int sum); \
unsigned int csumcpy_##a(const char *src, char *dst, \
                        int len, int sum, int *src_err_ptr, int *dst_err_ptr)

/* This makes adding/removing test functions easier
 */
/* asm ones... */

NAME(driver);

/* and C */
//#include "csumcpy_ntqpf.c"
//#include "csumcpy_ntqpf2.c"
//#include "csumcpy_ntqpfm.c"

#define TRY_TIMES 	8
#define BUFSZ   	(8 * 1024*1024)
#define CHUNKSZ 	(1 * 1024)
#define BUF_ALIGN 	1
#define MAX_START_OFS	512
#define MAX_END_OFS	512
#define POISON		1

#define CHUNKCNT (BUFSZ/2 / CHUNKSZ)

typedef unsigned char u8;

typedef unsigned int csum_func(const unsigned char * buff, int len,
		unsigned int sum);
typedef unsigned int csumcpy_func(const char *src, char *dst,
		int len, int sum, int *src_err_ptr, int *dst_err_ptr);

char *buffer;
char *chunk_addr[CHUNKCNT];
int start_ofs[CHUNKCNT];
int end_ofs[CHUNKCNT];
int chunk_sum[CHUNKCNT];

csum_func *best_csum;
csumcpy_func *best_csumcpy;

void kernel_fpu_begin(void) {}

static inline long long rdtsc()
{
	unsigned int low,high;
	__asm__ __volatile__("rdtsc" : "=a" (low), "=d" (high))
;
	return low + (((long long)high)<<32);
}

int die(const char *msg) {
	puts(msg);
	abort();
	return 1;
}

unsigned fold_csum(unsigned sum) {
	sum = (sum & 0xFFFF) + ((sum>>16) & 0xFFFF);
	if(sum >= 0x10000) sum = sum-0x10000+1;
	return sum;
}

void prepare() {
	int i;

	//ramdom buffer contents
        for(i=0; i<BUFSZ; i++) {
		buffer[i] = rand();
	}
	//linear chunk addresses...
        for(i=0; i<CHUNKCNT; i++) {
		chunk_addr[i] = buffer + CHUNKSZ*i;
	}
	//randomize 'em
        for(i=0; i<CHUNKCNT-1; i++) {
		int r = i+rand()%(CHUNKCNT-i);
		char *tp = chunk_addr[i];
		chunk_addr[i] = chunk_addr[r];
		chunk_addr[r] = tp;
	}
	//ramdom chunk offsets
        for(i=0; i<CHUNKCNT; i++) {
		// start must be even (really?)
		//start_ofs[i] = (rand() % (MAX_START_OFS/2+1))<<1;
		start_ofs[i] = rand() % (MAX_START_OFS+1);
		end_ofs[i] = rand() % (MAX_END_OFS+1);
		//start_ofs[i] = 2;
		//end_ofs[i] = 16;
	}
}

void dump(char *addr, int sz) {
	printf("\nbuf=0x%08x,sz=0x%05x\n",addr,sz);
	printf("addr    (+ofs ): byte\n");
	int i;
	for(i=-2; i<sz+2; i++) {
		printf("%08x(+%04x): %02x\n",
			addr+i, i, (u8)addr[i]
		);
	}
}

void dumpcpy(char *addr, int sz) {
	printf("\nbuf=0x%08x,sz=0x%05x\n",addr,sz);
	printf("addr    (+ofs ): org cpy\n");
	int i;
	for(i=-2; i<sz+2; i++) {
		printf("%08x(+%04x): %02x %02x%s\n",
			addr+i, i,
			(u8)addr[i], (u8)addr[i+BUFSZ/2],
			(u8)addr[i]==(u8)addr[i+BUFSZ/2] ? "" : " <"
		);
	}
}

unsigned test_one_csum(csum_func *func, char *name) {
	int i,j;
	unsigned long long before,after,min,max;
	unsigned init_sum,sum,sum2;

	printf("%-20s", name);
	best_csum = func;
	
	// used to catch 'small buffer' bugs
	init_sum = 0;
	init_sum = csum_partial(buffer,1,init_sum);
	init_sum = csum_partial(buffer,2,init_sum);
	init_sum = csum_partial(buffer,3,init_sum);
	init_sum = csum_partial(buffer+2,1,init_sum);
	init_sum = csum_partial(buffer+2,2,init_sum);
	init_sum = csum_partial(buffer+2,3,init_sum);

	// dummy run to ensure coherent mem/cache state
	sum = init_sum;
	for(j=0; j<CHUNKCNT; j++) {
		char *addr = chunk_addr[j]+start_ofs[j];
		int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
		if(sz<0) die("sz<0!");
		sum = csum_partial(addr,sz,sum);
		if(chunk_sum[j]) {
			if(chunk_sum[j]!=fold_csum(sum)) {
				dump(addr,sz);
				printf("sum:%04x expected:%04x\n",
					fold_csum(sum),chunk_sum[j]);
				die("bad sum!");
			}
		} else {
			chunk_sum[j] = fold_csum(sum);
		}
//debug printf("addr:0x%08x sz:%i\n",addr,sz);
	}

	// pick fastest run
	min = ~0ULL;
	max = 0;
	for (i=0;i<TRY_TIMES;i++) {
		sum2 = init_sum;
		before = rdtsc();
    		for(j=0; j<CHUNKCNT; j++) {
			char *addr = chunk_addr[j]+start_ofs[j];
			int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
			sum2 = csum_partial(addr,sz,sum2);
		}
		after = rdtsc();
		if(sum != sum2) die("bad sum");
		if(before > after) die("timer overflow");
		after -= before;
		if(min > after) min = after;
		if(max < after) max = after;
	}
	printf("%6lli max,%5lli min cycles/kb. sum=%08x=%04x\n",
		max / (BUFSZ/2 / 1024),
		min / (BUFSZ/2 / 1024),
		sum,
		fold_csum(sum)
	);
	return sum;
}

unsigned test_one_copy(csumcpy_func *func, char *name) {
	int i,j;
	unsigned long long before,after,min,max;
	unsigned init_sum,sum,sum2;
	int err;

	printf("%-20s", name);
	best_csumcpy = func;
	
	// used to catch 'small buffer' bugs
	init_sum = 0;
	init_sum = csum_partial_copy_generic(buffer,  buffer  +BUFSZ/2,1,init_sum,&err,&err);
	init_sum = csum_partial_copy_generic(buffer,  buffer  +BUFSZ/2,2,init_sum,&err,&err);
	init_sum = csum_partial_copy_generic(buffer,  buffer  +BUFSZ/2,3,init_sum,&err,&err);
	init_sum = csum_partial_copy_generic(buffer+2,buffer+2+BUFSZ/2,1,init_sum,&err,&err);
	init_sum = csum_partial_copy_generic(buffer+2,buffer+2+BUFSZ/2,2,init_sum,&err,&err);
	init_sum = csum_partial_copy_generic(buffer+2,buffer+2+BUFSZ/2,3,init_sum,&err,&err);

	// dummy run to ensure coherent mem/cache state
	if(POISON) memset(buffer+BUFSZ/2,0xaa,BUFSZ/2);
	sum = init_sum;
	for(j=0; j<CHUNKCNT; j++) {
		char *addr = chunk_addr[j]+start_ofs[j];
		int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
		sum = csum_partial_copy_generic(addr,addr+BUFSZ/2,sz,sum,&err,&err);
		if(POISON) {
			if(memcmp(addr,addr+BUFSZ/2,sz)!=0) {
				dumpcpy(addr,sz);
				die("bad copy");
			}
			if(start_ofs[j] && addr[BUFSZ/2-1]!=(char)0xaa) {
				dumpcpy(addr,sz);
				die("error at start");
			}
			if(end_ofs[j] && addr[BUFSZ/2+sz]!=(char)0xaa) {
				dumpcpy(addr,sz);
				die("error at end");
			}
		}
		if(chunk_sum[j]) {
			if(chunk_sum[j]!=fold_csum(sum)) {
				dumpcpy(addr,sz);
				printf("sum:%04x expected:%04x\n",
					fold_csum(sum),chunk_sum[j]);
				die("bad sum!");
			}
		} else {
			chunk_sum[j] = fold_csum(sum);
		}
	}

	// pick fastest run
	min = ~0ULL;
	max = 0;
	for (i=0; i<TRY_TIMES; i++) {
		sum2 = init_sum;
		before = rdtsc();
		for(j=0; j<CHUNKCNT; j++) {
			char *addr = chunk_addr[j]+start_ofs[j];
			int sz = CHUNKSZ-start_ofs[j]-end_ofs[j];
			sum2 = csum_partial_copy_generic(addr,addr+BUFSZ/2,sz,sum2,&err,&err);
		}
		after = rdtsc();
		if(sum != sum2) die("bad sum");
		if(before>after) die("timer overflow");
		after -= before;
		if(min > after) min = after;
		if(max < after) max = after;
	}

	printf("%6lli max,%5lli min cycles/kb. sum=%08x=%04x\n",
		max / (BUFSZ/2 / 1024),
		min / (BUFSZ/2 / 1024),
		sum,
		fold_csum(sum)
	);
	return sum;
}

NAME(naive);
NAME(basic);
NAME(3dnow);
NAME(ssemmxplus);
NAME(sse);

void test_csum()
{
	unsigned sum;
	puts("\tcsum tests:");

#define	TEST_CSUM(a) fold_csum(test_one_csum(a,#a))
	sum =  TEST_CSUM(csum_basic	);
	sum == TEST_CSUM(csum_basic	) || die("bad sum");
	sum == TEST_CSUM(csum_basic	) || die("bad sum");
	sum == TEST_CSUM(csum_naive	) || die("bad sum");
	sum == TEST_CSUM(csum_3dnow	) || die("bad sum");
	sum == TEST_CSUM(csum_ssemmxplus) || die("bad sum");
//	sum == TEST_CSUM(csum_sse	) || die("bad sum");
#undef TEST_CSUM
}

void test_copy()
{
	unsigned sum;
	puts("\tcopy tests:");

#define	TEST_COPY(a) fold_csum(test_one_copy(a,#a))
	sum =  TEST_COPY(csumcpy_basic		);
	sum == TEST_COPY(csumcpy_basic		) || die("bad sum");
	sum == TEST_COPY(csumcpy_basic		) || die("bad sum");
	sum == TEST_COPY(csumcpy_naive		) || die("bad sum");
	sum == TEST_COPY(csumcpy_ssemmxplus	) || die("bad sum");
//	sum == TEST_COPY(csumcpy_sse		) || die("bad sum");
#undef TEST_COPY
}

int main()
{
	char *buffer_raw;
//	time_t seed = 1036108002;
//	time_t seed = 1036104278;
	time_t seed = time(0);
	srand(seed);
	
	printf("Random seed: %i  Buffer: %i Mb (%i)  Chunks: %i kb (%i)\n"
		"Start/end offset in chunk: random from 0 to %i/%i\n"
		"Each test tried %i times, max and min CPU cycles are reported\n"
		"Please disregard max values, they are due to system interference only\n",
		seed,
		BUFSZ/1024/1024, BUFSZ,
		CHUNKSZ/1024, CHUNKSZ,
		MAX_START_OFS,MAX_END_OFS,
		TRY_TIMES
	);


	// +31+ and +rand()%32+ help reach 4-unaligned addresses
	// (glibc malloc can do some alignment)
        buffer_raw = malloc(BUFSZ+31+BUF_ALIGN-1);
        if(!buffer_raw) die("Malloc failed");
        buffer = (char*) ((((int)buffer_raw)+rand()%32+(BUF_ALIGN-1)) & (~(BUF_ALIGN-1)));

	prepare();
	test_csum();
	test_copy();

	puts("Done");
	free(buffer_raw);
	return 0;
}
